/*
 * Copyright (C) 2010 Ajax.org BV
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this library; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#ifndef J_C_XML_ELEMENT2_H
#define J_C_XML_ELEMENT2_H


namespace o3 {
    struct cXmlElement : cXmlNode, iXmlElement2 {
        cXmlElement(xmlNodePtr node, iXmlNode2* owner_node, NodeMap* node_map) 
            : cXmlNode(node, owner_node, node_map) {
            o3_trace_scrfun("cXmlElement");
        }

		virtual ~cXmlElement()
		{

		}

        o3_begin_class(cXmlNode)
            o3_add_iface(iXmlElement2)
        o3_end_class();

        o3_glue_gen()
        
        o3_fun siXmlNode2 selectSingleNode(iCtx* ctx, const char* selection, iScr* ctxt = 0) {
            o3_trace_scrfun("selectSingleNode");
            xmlXPathObjectPtr res = selectNodesInternal(
                selection, siXmlDocument2(ctxt), ctx);
            if (!res || !res->nodesetval || !res->nodesetval->nodeTab) 
                return siXmlNode2();
            siXmlNode2 ret = wrapNode(ctx, res->nodesetval->nodeTab[0], m_owner_node ? m_owner_node : this);
            xmlXPathFreeObject(res);
            return ret;
        }

        o3_fun siXmlNodeArray2 selectNodes(iCtx* ctx, const char* selection, iScr* ctxt = 0) {
            o3_trace_scrfun("selectNodes");
            xmlXPathObjectPtr res = selectNodesInternal(
                selection, siXmlDocument2(ctxt), ctx);
            if (!res || !res->nodesetval)
                return o3_new(cXmlNodeArray)(this);
            cXmlNodeArray* nodes = o3_new(cXmlNodeArray)(m_owner_node ? m_owner_node : this);
            nodes->m_nodes = res->nodesetval;
            nodes->m_xpath_obj = res;
            return siXmlNodeArray2(nodes);            
        }
        
        virtual o3_get Str tagName() {
            o3_trace_scrfun("tagName");
            return Str((const char*) m_node->name);
        }
        
        virtual o3_fun Str getAttribute(iCtx* ctx, const char* name) {
			o3_trace_scrfun("getAttribute");
			o3_assert(name);
			siXmlAttr2 attr = getAttributeNode(ctx, name);
			if (attr)
				return attr->value();

			if (strCompare("xmlns", name, 5))
				return Str();
			
			const char* pre = *(name+5) == ':' ? name+6 : 0;

			xmlNsPtr ns = m_node->nsDef;
			while (ns) {
				if (!ns->prefix && !pre)
					return Str((const char*) ns->href);
				
				if (ns->prefix && pre && !strCompare((const char*) ns->prefix, pre)) 
					return Str((const char*) ns->href);
				
				ns = ns->next;
			}

			return Str();
        }
        
        virtual o3_fun void setAttribute(iCtx* ctx, const char* name, const char* value) {
            o3_trace_scrfun("setAttribute");
            o3_assert(name);
            o3_assert(value);
            siXmlAttr2 new_attr;
			if (ownerDocument(ctx))
				new_attr = ownerDocument(ctx)->createAttribute(ctx, name);
            else{
				xmlAttrPtr attr = xmlNewDocProp(0, BAD_CAST name, 0);
				new_attr = wrapNode(ctx, (xmlNodePtr) attr, 0);
			}
						
			new_attr->setValue(value);
            setAttributeNode(ctx, new_attr);
        }
        
        virtual o3_fun void removeAttribute(iCtx* ctx, const char* name) {
            o3_trace_scrfun("removeAttribute");
            o3_assert(name);
            removeAttributeNode(getAttributeNode(ctx, name));
        }
        
        o3_fun siXmlAttr2 getAttributeNode(iCtx* ctx, const char* name) {
            o3_trace_scrfun("getAttributeNode");
            o3_assert(name);
            xmlAttrPtr attr = xmlHasProp(m_node, BAD_CAST name);
            
            if (attr == NULL)
                return siXmlAttr2();
            return wrapNode(ctx, (xmlNodePtr) attr, m_owner_node ? m_owner_node : this);
        }
        
        o3_fun siXmlAttr2 setAttributeNode(iCtx* ctx, iXmlAttr2* new_attr, siEx* ex = 0) {
            o3_trace_scrfun("setAttributeNode");            
            return new_attr ? appendChild(ctx, siXmlNode2(new_attr), ex) : siXmlNode2();
        }
        
        o3_fun siXmlAttr2 removeAttributeNode(iXmlAttr2* old_attr) {
            o3_trace_scrfun("removeAttributeNode");                
            return old_attr ? removeChild(siXmlNode2(old_attr)) : siXmlNode2();
        }
        
		o3_fun siXmlNodeList2 getElementsByTagName(const char* name) {
            o3_trace_scrfun("getElementsByTagName");
            o3_assert(name);
            return siXmlNodeList2(o3_new(cXmlNodeList)(this, name));
		}

        o3_fun siXmlElement2 getElementById(iCtx* ctx, const char* id) {
            o3_trace_scrfun("getElementById");
            if (strEquals(getAttribute(ctx, "id").ptr(), id))
                return this;
            for (siXmlNode2 child = firstChild(ctx); child; child = child->nextSibling(ctx))
                if (siXmlElement2 element = child)
                    if (element = element->getElementById(ctx, id))
                        return element;
            return siXmlElement2();
        }
        
        xmlXPathObjectPtr selectNodesInternal(const char* expr, iXmlDocument2* ctxt, iCtx* ctx) {
            o3_trace_scrfun("selectNodesInternal");
            o3_assert(expr);

            xmlXPathObjectPtr xpathObj;
            xmlXPathContextPtr xpathCtx;             
            // Create xpath evaluation context
            if (ctxt){
                //search from the doc
                xpathCtx = xmlXPathNewContext(ctxt->docPtr());
                o3_assert(xpathCtx);
            } else {
                //search from current node
                o3_assert(m_node);
                xpathCtx = xmlXPathNewContext( m_node->doc );
                o3_assert(xpathCtx);
                xpathCtx->node = m_node;                
            }
            //registering namespaces
            siXmlDocument2 doc = ownerDocument(ctx);
			if (doc){				
				const tMap<Str,Str> ns = doc->nameSpaces();
				tMap<Str,Str>::ConstIter it;
				for (it=ns.begin(); it != ns.end(); ++it) {
					if (xmlXPathRegisterNs(xpathCtx, 
						BAD_CAST (*it).key.ptr(),
						BAD_CAST (*it).val.ptr()))
					{
						o3_assert(false);
					}
				}
            }
            //do the xpath eval
            xpathObj = xmlXPathEval(BAD_CAST expr, xpathCtx);
            if(!xpathObj) {
                xmlXPathFreeContext(xpathCtx); 
                return 0;
            }
            
            xmlXPathFreeContext(xpathCtx); 
            return xpathObj;            
        }
        
        o3_fun void normalize(iCtx* ctx) {
            o3_trace_scrfun("normalize");

            siXmlNode2 child = firstChild(ctx);
            while (child != NULL) {
                if (child->nodeType() == TYPE_TEXT) {
                    Str value;

                    do {
                        siXmlNode2 sibling = child->nextSibling(ctx); 

                        value += child->nodeValue();
                        removeChild(child);
                        child = sibling;
                    } while (child != NULL && child->nodeType() == TYPE_TEXT);
                    if (!value.empty()) {
                        siXmlText2 text = ownerDocument(ctx)->createTextNode(
                                ctx, value); 

                        insertBefore(ctx, siXmlNode2(text).ptr(), child);
                    }
                } else {
                    if (siXmlElement2 element = child) 
                        element->normalize(ctx);
                    child = child->nextSibling(ctx);
                } 
            }
        }
        
        virtual xmlAttrPtr firstAttr() {
            o3_trace_scrfun("firstAttr");
            return m_node->properties;
        }
    };
}

#endif // J_C_XML_ELEMENT_H
